@@ -7,8 +7,175 @@ use warnings;
 
 use Exporter qw(import);
 
-our $VERSION = '2.20';
-our @EXPORT  = qw(purl_to_urls);
+our $VERSION = '2.21';
+our @EXPORT  = qw(purl_to_urls purl_components_normalize);
+
+sub purl_components_normalize {
+
+    my (%component) = @_;
+
+    my %TYPES = (
+        conan       => \&_conan_normalize,
+        cpan        => \&_cpan_normalize,
+        cran        => \&_cran_normalize,
+        huggingface => \&_huggingface_normalize,
+        mlflow      => \&_mlflow_normalize,
+        pypi        => \&_pypi_normalize,
+        swid        => \&_swid_normalize,
+        swift       => \&_swift_normalize,
+    );
+
+    Carp::croak "Invalid Package URL: '$component{scheme}' is not a valid scheme" unless ($component{scheme} eq 'pkg');
+
+    $component{type} = lc $component{type};
+
+    if (grep { $_ eq $component{type} } qw(alpm apk bitbucket composer deb github gitlab hex npm oci pypi)) {
+        $component{name} = lc $component{name};
+    }
+
+    if (defined $component{namespace}) {
+        if (grep { $_ eq $component{type} } qw(alpm apk bitbucket composer deb github gitlab golang hex rpm)) {
+            $component{namespace} = lc $component{namespace};
+        }
+    }
+
+    foreach my $qualifier (keys %{$component{qualifiers}}) {
+        Carp::croak "Invalid Package URL: '$qualifier' is not a valid qualifier" if ($qualifier =~ /\s/);
+        Carp::croak "Invalid Package URL: '$qualifier' is not a valid qualifier" if ($qualifier =~ /\%/);
+    }
+
+    if (defined $TYPES{$component{type}}) {
+        return $TYPES{$component{type}}->(%component);
+    }
+
+    return \%component;
+
+}
+
+sub _conan_normalize {
+
+    my (%component) = @_;
+
+    if (defined $component{namespace} && $component{namespace} ne '') {
+        if (!defined $component{qualifiers}->{channel}) {
+            Carp::croak
+                "Invalid Package URL: Conan 'channel' qualifier does not exist for namespace '$component{namespace}'";
+        }
+    }
+    else {
+        if (defined $component{qualifiers}->{channel}) {
+            Carp::croak
+                "Invalid Package URL: Conan 'namespace' does not exist for channel '$component{qualifiers}->{channel}'";
+        }
+    }
+
+    return \%component;
+
+}
+
+sub _cpan_normalize {
+
+    my (%component) = @_;
+
+    # To refer to a CPAN distribution name, the "namespace" MUST be present. In this
+    # case, the "namespace" is the CPAN id of the author/publisher. It MUST be
+    # written uppercase, followed by the distribution name in the "name" component. A
+    # distribution name MUST NOT contain the string "::".
+
+    # To refer to a CPAN module, the "namespace" MUST be absent. The module name MAY
+    # contain zero or more "::" strings, and the module name MUST NOT contain a "-"
+
+    $component{namespace} = uc $component{namespace} if (defined $component{namespace});
+
+    if ((defined $component{namespace} && defined $component{name}) && $component{namespace} =~ /\:/) {
+        Carp::croak "Invalid Package URL: CPAN 'namespace' component must have the distribution author";
+    }
+
+    if ((defined $component{namespace} && defined $component{name}) && $component{name} =~ /\:/) {
+        Carp::croak "Invalid Package URL: CPAN 'name' component must have the distribution name";
+    }
+
+    if (!defined $component{namespace} && $component{name} =~ /\-/) {
+        Carp::croak "Invalid Package URL: CPAN 'name' component must have the module name";
+    }
+
+    return \%component;
+
+}
+
+sub _cran_normalize {
+
+    my (%component) = @_;
+
+    Carp::croak "Invalid Package URL: Cran 'version' is required" unless defined $component{version};
+
+    return \%component;
+
+}
+
+sub _huggingface_normalize {
+
+    my (%component) = @_;
+
+    # The version is the model revision Git commit hash. It is case insensitive and
+    # must be lowercased in the package URL.
+    $component{version} = lc $component{version};
+
+    return \%component;
+
+}
+
+sub _mlflow_normalize {
+
+    my (%component) = @_;
+
+    # The "name" case sensitivity depends on the server implementation:
+    #   - Azure ML: it is case sensitive and must be kept as-is in the package URL.
+    #   - Databricks: it is case insensitive and must be lowercased in the package URL.
+
+    if (defined $component{qualifiers}->{repository_url}
+        && $component{qualifiers}->{repository_url} =~ /azuredatabricks/)
+    {
+        $component{name} = lc $component{name};
+    }
+
+    return \%component;
+
+}
+
+sub _pypi_normalize {
+
+    my (%component) = @_;
+
+    # A PyPI package name must be lowercased and underscore "_" replaced with a dash "-".
+    $component{name} =~ s/_/-/g;
+
+    return \%component;
+
+}
+
+sub _swid_normalize {
+
+    my (%component) = @_;
+
+    Carp::croak "Invalid Package URL: swid 'tag_id' qualifier is required"
+        unless defined $component{qualifiers}->{tag_id};
+
+    return \%component;
+
+}
+
+sub _swift_normalize {
+
+    my (%component) = @_;
+
+    Carp::croak "Invalid Package URL: Swift 'version' is required"   unless defined $component{version};
+    Carp::croak "Invalid Package URL: Swift 'namespace' is required" unless defined $component{namespace};
+
+    return \%component;
+
+}
+
 
 sub purl_to_urls {
 
@@ -50,61 +217,24 @@ sub purl_to_urls {
 
 }
 
-sub _github_urls {
+sub _bitbucket_urls {
 
     my $purl = shift;
 
-    my $name           = $purl->name;
-    my $namespace      = $purl->namespace;
-    my $version        = $purl->version;
-    my $qualifiers     = $purl->qualifiers;
-    my $file_ext       = $qualifiers->{ext}            || 'tar.gz';
-    my $version_prefix = $qualifiers->{version_prefix} || '';
+    my $name       = $purl->name;
+    my $namespace  = $purl->namespace;
+    my $version    = $purl->version;
+    my $qualifiers = $purl->qualifiers;
+    my $file_ext   = $qualifiers->{ext} || 'tar.gz';
 
     my $urls = {};
 
     if ($name && $namespace) {
-        $urls->{repository} = "https://github.com/$namespace/$name";
+        $urls->{repository} = "https://bitbucket.org/$namespace/$name";
     }
 
     if ($version) {
-
-        my $is_sha1 = ($version =~ /^[a-fA-F0-9]{40}$/);
-
-        if ($is_sha1) {
-            $urls->{download} = "https://github.com/$namespace/$name/archive/$version.$file_ext";
-        }
-        else {
-            $urls->{download}
-                = "https://github.com/$namespace/$name/archive/refs/tags/$version_prefix$version.$file_ext";
-        }
-
-    }
-
-    return $urls;
-
-}
-
-sub _gitlab_urls {
-
-    my $purl = shift;
-
-    my $name           = $purl->name;
-    my $namespace      = $purl->namespace;
-    my $version        = $purl->version;
-    my $qualifiers     = $purl->qualifiers;
-    my $file_ext       = $qualifiers->{ext}            || 'tar.gz';
-    my $version_prefix = $qualifiers->{version_prefix} || '';
-
-    my $urls = {};
-
-    if ($name && $namespace) {
-        $urls->{repository} = "https://gitlab.com/$namespace/$name";
-    }
-
-    if ($version) {
-        $urls->{download}
-            = "https://gitlab.com/$namespace/$name/-/archive/$version_prefix$version/$name-$version_prefix$version.$file_ext";
+        $urls->{download} = "https://bitbucket.org/$namespace/$name/get/$version.$file_ext";
     }
 
     return $urls;
@@ -129,66 +259,16 @@ sub _cargo_urls {
 
 }
 
-sub _gem_urls {
+sub _composer_urls {
 
     my $purl = shift;
 
-    my $name    = $purl->name;
-    my $version = $purl->version;
-
-    if ($name && $version) {
-        return {
-            repository => "https://rubygems.org/gems/$name/versions/$version",
-            download   => "https://rubygems.org/downloads/$name-$version.gem"
-        };
-    }
-
-    return {repository => "https://rubygems.org/gems/$name"};
-
-}
-
-sub _pypi_urls {
-
-    my $purl = shift;
-
-    my $name    = $purl->name;
-    my $version = $purl->version;
-
-    if ($name && $version) {
-        return {repository => "https://pypi.org/project/$name/$version"};
-    }
-
-    return {repository => "https://pypi.org/project/$name"};
-
-}
-
-sub _npm_urls {
-
-    my $purl = shift;
-
-    my $namespace = $purl->namespace;
     my $name      = $purl->name;
-    my $version   = $purl->version;
+    my $namespace = $purl->namespace;
 
-    if ($namespace && $name && $version) {
-        return {
-            repository => "https://www.npmjs.com/package/$namespace/$name/v/$version",
-            download   => "https://registry.npmjs.org/$namespace/$name/-/$name-$version.tgz"
-        };
-    }
+    return unless ($name && $namespace);
 
-    if ($name && $version) {
-        return {
-            repository => "https://www.npmjs.com/package/$name/v/$version",
-            download   => "https://registry.npmjs.org/$name/-/$name-$version.tgz"
-        };
-    }
-
-    if ($namespace && $name) {
-        return {repository => "https://www.npmjs.com/package/$namespace/$name"};
-    }
-
-    return {repository => "https://www.npmjs.com/package/$name"};
+    return {repository => "https://packagist.org/packages/$namespace/$name"};
 
 }
 
@@ -225,90 +305,6 @@ sub _cpan_urls {
 
 }
 
-sub _nuget_urls {
-
-    my $purl = shift;
-
-    my $name    = $purl->name;
-    my $version = $purl->version;
-
-    if ($name && $version) {
-        return {
-            repository => "https://www.nuget.org/packages/$name/$version",
-            download   => "https://www.nuget.org/api/v2/package/$name/$version"
-        };
-    }
-
-    return {repository => "https://www.nuget.org/packages/$name"};
-
-}
-
-sub _maven_urls {
-
-    my $purl = shift;
-
-    my $namespace  = $purl->namespace;
-    my $name       = $purl->name;
-    my $version    = $purl->version;
-    my $qualifiers = $purl->qualifiers;
-    my $extension  = $qualifiers->{extension}      // 'jar';
-    my $repo_url   = $qualifiers->{repository_url} // 'repo1.maven.org/maven2';
-
-    if ($namespace && $name && $version) {
-
-        (my $ns_url = $namespace) =~ s/\./\//g;
-
-        return {
-            repository => "https://mvnrepository.com/artifact/$namespace/$name/$version",
-            download   => "https://$repo_url/$ns_url/$name/$version/$name-$version.$extension"
-        };
-
-    }
-
-    if ($namespace && $name) {
-        return {repository => "https://mvnrepository.com/artifact/$namespace/$name"};
-    }
-
-}
-
-sub _composer_urls {
-
-    my $purl = shift;
-
-    my $name      = $purl->name;
-    my $namespace = $purl->namespace;
-
-    return unless ($name && $namespace);
-
-    return {repository => "https://packagist.org/packages/$namespace/$name"};
-
-}
-
-sub _bitbucket_urls {
-
-    my $purl = shift;
-
-    my $name           = $purl->name;
-    my $namespace      = $purl->namespace;
-    my $version        = $purl->version;
-    my $qualifiers     = $purl->qualifiers;
-    my $file_ext       = $qualifiers->{ext}            || 'tar.gz';
-    my $version_prefix = $qualifiers->{version_prefix} || '';
-
-    my $urls = {};
-
-    if ($name && $namespace) {
-        $urls->{repository} = "https://bitbucket.org/$namespace/$name";
-    }
-
-    if ($version) {
-        $urls->{download} = "https://bitbucket.org/$namespace/$name/get/$version_prefix$version.$file_ext";
-    }
-
-    return $urls;
-
-}
-
 sub _docker_urls {
 
     my $purl = shift;
@@ -341,6 +337,81 @@ sub _docker_urls {
 
 }
 
+sub _gem_urls {
+
+    my $purl = shift;
+
+    my $name    = $purl->name;
+    my $version = $purl->version;
+
+    if ($name && $version) {
+        return {
+            repository => "https://rubygems.org/gems/$name/versions/$version",
+            download   => "https://rubygems.org/downloads/$name-$version.gem"
+        };
+    }
+
+    return {repository => "https://rubygems.org/gems/$name"};
+
+}
+
+sub _github_urls {
+
+    my $purl = shift;
+
+    my $name       = $purl->name;
+    my $namespace  = $purl->namespace;
+    my $version    = $purl->version;
+    my $qualifiers = $purl->qualifiers;
+    my $file_ext   = $qualifiers->{ext} || 'tar.gz';
+
+    my $urls = {};
+
+    if ($name && $namespace) {
+        $urls->{repository} = "https://github.com/$namespace/$name";
+    }
+
+    if ($version) {
+
+        my $is_sha1 = ($version =~ /^[a-fA-F0-9]{40}$/);
+
+        if ($is_sha1) {
+            $urls->{download} = "https://github.com/$namespace/$name/archive/$version.$file_ext";
+        }
+        else {
+            $urls->{download} = "https://github.com/$namespace/$name/archive/refs/tags/$version.$file_ext";
+        }
+
+    }
+
+    return $urls;
+
+}
+
+sub _gitlab_urls {
+
+    my $purl = shift;
+
+    my $name       = $purl->name;
+    my $namespace  = $purl->namespace;
+    my $version    = $purl->version;
+    my $qualifiers = $purl->qualifiers;
+    my $file_ext   = $qualifiers->{ext} || 'tar.gz';
+
+    my $urls = {};
+
+    if ($name && $namespace) {
+        $urls->{repository} = "https://gitlab.com/$namespace/$name";
+    }
+
+    if ($version) {
+        $urls->{download} = "https://gitlab.com/$namespace/$name/-/archive/$version/$name-$version.$file_ext";
+    }
+
+    return $urls;
+
+}
+
 sub _golang_urls {
 
     my $purl = shift;
@@ -392,6 +463,97 @@ sub _luarocks_urls {
 
 }
 
+sub _maven_urls {
+
+    my $purl = shift;
+
+    my $namespace  = $purl->namespace;
+    my $name       = $purl->name;
+    my $version    = $purl->version;
+    my $qualifiers = $purl->qualifiers;
+    my $extension  = $qualifiers->{extension}      // 'jar';
+    my $repo_url   = $qualifiers->{repository_url} // 'repo1.maven.org/maven2';
+
+    if ($namespace && $name && $version) {
+
+        (my $ns_url = $namespace) =~ s/\./\//g;
+
+        return {
+            repository => "https://mvnrepository.com/artifact/$namespace/$name/$version",
+            download   => "https://$repo_url/$ns_url/$name/$version/$name-$version.$extension"
+        };
+
+    }
+
+    if ($namespace && $name) {
+        return {repository => "https://mvnrepository.com/artifact/$namespace/$name"};
+    }
+
+}
+
+sub _npm_urls {
+
+    my $purl = shift;
+
+    my $namespace = $purl->namespace;
+    my $name      = $purl->name;
+    my $version   = $purl->version;
+
+    if ($namespace && $name && $version) {
+        return {
+            repository => "https://www.npmjs.com/package/$namespace/$name/v/$version",
+            download   => "https://registry.npmjs.org/$namespace/$name/-/$name-$version.tgz"
+        };
+    }
+
+    if ($name && $version) {
+        return {
+            repository => "https://www.npmjs.com/package/$name/v/$version",
+            download   => "https://registry.npmjs.org/$name/-/$name-$version.tgz"
+        };
+    }
+
+    if ($namespace && $name) {
+        return {repository => "https://www.npmjs.com/package/$namespace/$name"};
+    }
+
+    return {repository => "https://www.npmjs.com/package/$name"};
+
+}
+
+sub _nuget_urls {
+
+    my $purl = shift;
+
+    my $name    = $purl->name;
+    my $version = $purl->version;
+
+    if ($name && $version) {
+        return {
+            repository => "https://www.nuget.org/packages/$name/$version",
+            download   => "https://www.nuget.org/api/v2/package/$name/$version"
+        };
+    }
+
+    return {repository => "https://www.nuget.org/packages/$name"};
+
+}
+
+sub _pypi_urls {
+
+    my $purl = shift;
+
+    my $name    = $purl->name;
+    my $version = $purl->version;
+
+    if ($name && $version) {
+        return {repository => "https://pypi.org/project/$name/$version"};
+    }
+
+    return {repository => "https://pypi.org/project/$name"};
+
+}
+
 1;
 
 __END__
@@ -403,7 +565,7 @@ URI::PackageURL::Util - Utility for URI::PackageURL
 
   use URI::PackageURL::Util qw(purl_to_urls);
 
-  $urls = purl_to_urls('pkg:cpan/GDT/URI-PackageURL@2.01');
+  $urls = purl_to_urls('pkg:cpan/GDT/URI-PackageURL@2.20');
 
   $filename = basename($urls->{download});
   $ua->mirror($urls->{download}, "/tmp/$filename");
@@ -415,7 +577,11 @@ URL::PackageURL::Util is the utility package for URL::PackageURL.
 
 =over
 
-=item $urls = purl_to_urls($purl_string | URI::PackageURL);
+=item %normalized_purl_components = purl_components_normalize(%purl_components)
+
+Normalize the given Package URL components
+
+=item $urls = purl_to_urls($purl_string | URI::PackageURL)
 
 Converts the given Package URL string or L<URI::PackageURL> instance and return
 the hash with C<repository> and/or C<download> URL.
